# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Main entrypoint for running transport simulation.

Example command with a configuration defined in Python:
python3 run_simulation_main.py \
 --config=examples/iterhybrid_rampup.py \
 --log_progress
"""

from collections.abc import Sequence
import enum
import functools
import os
import pathlib
import time

from absl import app
from absl import flags
from absl import logging
import jax
import torax
from torax._src import simulation_app
from torax._src.config import config_loader
from torax._src.plotting import plotruns_lib
from torax._src.torax_pydantic import model_config

os.environ['XLA_FLAGS'] = (
    os.environ.get('XLA_FLAGS', '')
    + ' --xla_backend_extra_options=xla_cpu_flatten_after_fusion'
)

# String used when prompting the user to make a choice of command
CHOICE_PROMPT = 'Your choice: '
# String used when prompting the user to make a yes / no choice
Y_N_PROMPT = 'y/n: '
# String used when printing how long the simulation took
SIMULATION_TIME = 'simulation time'

FLAGS = flags.FLAGS

_CONFIG_PATH = flags.DEFINE_string(
    'config',
    None,
    'The absolute or relative path to the Python file to import the config'
    ' from. This file must contain a global variable named `CONFIG`'
    ' representing the config dictionary. If it is a relative path, it will'
    ' first be attempted to be resolved relative to the working directory, and'
    ' then the Torax base directory.',
)

_LOG_SIM_PROGRESS = flags.DEFINE_bool(
    'log_progress',
    False,
    'If true, logs the time of each timestep as the simulation runs.',
)

_PLOT_SIM_PROGRESS = flags.DEFINE_bool(
    'plot_progress',
    False,
    'If true, plots the time of each timestep as the simulation runs.'
    ' Note: this is temporarily disabled.',
)

_LOG_SIM_OUTPUT = flags.DEFINE_bool(
    'log_output',
    False,
    'If True, logs extra information to stdout/stderr.',
)

_REFERENCE_RUN = flags.DEFINE_string(
    'reference_run',
    None,
    'If provided, after the simulation is run, we can compare the last run to'
    ' the reference run.',
)

_USE_JAX_PROFILER = flags.DEFINE_bool(
    'use_jax_profiler',
    False,
    'If True, use Jax profiler. Processing will stop and a URL to visualise'
    ' results will be generated. Processing will resume when the URL is opened.'
    'See https://jax.readthedocs.io/en/latest/profiling.html.',
)

_QUIT = flags.DEFINE_bool(
    'quit',
    False,
    'If True, quits after the first operation (no interactive mode).',
)

# Needed to test-time name collision with the flag in regenerate_torax_refs.py.
if 'output_dir' not in FLAGS:
  _OUTPUT_DIR = flags.DEFINE_string(
      'output_dir',
      None,
      'If provided, overrides the default output directory.',
  )
else:
  _OUTPUT_DIR = None

_PLOT_CONFIG_PATH = flags.DEFINE_string(
    'plot_config',
    'plotting/configs/default_plot_config.py',
    'Path to a Python file containing a `PLOT_CONFIG` variable. If it is a'
    ' relative path, it will first be attempted to be resolved relative to the'
    ' working directory, and then the Torax base directory.',
)

jax.config.parse_flags_with_absl()


@enum.unique
class _UserCommand(enum.Enum):
  """Options to do on every iteration of the script."""

  RUN = ('RUN SIMULATION', 'r', simulation_app.AnsiColors.GREEN)
  MODIFY_CONFIG = ('modify the existing config and reload it', 'mc')
  CHANGE_CONFIG = ('provide a new config file to load', 'cc')
  TOGGLE_LOG_SIM_PROGRESS = ('toggle --log_progress', 'tlp')
  TOGGLE_LOG_SIM_OUTPUT = ('toggle --log_output', 'tlo')
  PLOT_RUN = ('plot previous run(s) or against reference if provided', 'pr')
  QUIT = ('quit', 'q', simulation_app.AnsiColors.RED)


def _prompt_user(config_path: pathlib.Path) -> _UserCommand:
  """Prompts the user for the next thing to do."""
  simulation_app.log_to_stdout('\n')
  simulation_app.log_to_stdout(
      f'Using the config: {config_path}',
      color=simulation_app.AnsiColors.YELLOW,
  )
  user_command = None
  simulation_app.log_to_stdout('\n')
  while user_command is None:
    simulation_app.log_to_stdout(
        'What would you like to do next?',
        color=simulation_app.AnsiColors.BLUE,
    )
    for uc in _UserCommand:
      if len(uc.value) == 3:
        color = uc.value[2]
      else:
        color = simulation_app.AnsiColors.BLUE
      simulation_app.log_to_stdout(f'{uc.value[1]}: {uc.value[0]}', color)
    input_text = input(CHOICE_PROMPT)
    text_to_uc = {uc.value[1]: uc for uc in _UserCommand}
    input_text = input_text.lower().strip()
    if input_text in text_to_uc:
      user_command = text_to_uc[input_text]
    else:
      simulation_app.log_to_stdout(
          'Unrecognized input. Try again.',
          color=simulation_app.AnsiColors.YELLOW,
      )
  return user_command


def _maybe_update_config(path: pathlib.Path) -> pathlib.Path:
  """Returns a possibly-updated config module to import."""
  simulation_app.log_to_stdout(
      f'Existing config module path: {path}',
      color=simulation_app.AnsiColors.BLUE,
  )
  simulation_app.log_to_stdout(
      'Would you like to change which file to import?',
      color=simulation_app.AnsiColors.BLUE,
  )
  should_change = _get_yes_or_no()
  if should_change:
    logging.info('Updating the config file path.')
    new_path = input('Enter the new path to use: ').strip()
    return pathlib.Path(new_path).resolve()
  else:
    logging.info('Continuing with %s', str(path))
    return path


def _modify_config(
    path: pathlib.Path,
) -> model_config.ToraxConfig | None:
  """Returns a new ToraxConfig from the modified config module."""

  simulation_app.log_to_stdout(
      f'Change {path} to include new values.',
      color=simulation_app.AnsiColors.BLUE,
  )
  simulation_app.log_to_stdout(
      'Modify the config with new values, then enter "y".',
      color=simulation_app.AnsiColors.BLUE,
  )
  simulation_app.log_to_stdout(
      'Enter "n" to go back to the main menu without loading any changes.',
      color=simulation_app.AnsiColors.BLUE,
  )
  proceed_with_run = _get_yes_or_no()

  if not proceed_with_run:
    return None

  return config_loader.build_torax_config_from_file(path)


def _change_config(
    config_path: pathlib.Path,
) -> tuple[model_config.ToraxConfig, pathlib.Path]:
  """Returns a ToraxConfig from the new config file."""
  config_path = _maybe_update_config(config_path)
  simulation_app.log_to_stdout(
      f'Change {config_path} to include new values. Any changes to '
      'CONFIG will be picked up.',
      color=simulation_app.AnsiColors.BLUE,
  )
  input('Press Enter when done changing the module.')
  torax_config = config_loader.build_torax_config_from_file(config_path)
  return torax_config, config_path


def _get_yes_or_no() -> bool:
  """Returns a boolean indicating yes depending on user input."""
  input_text = None
  while input_text is None:
    input_text = input(Y_N_PROMPT)
    input_text = input_text.lower().strip()
    if input_text not in ('y', 'n'):
      simulation_app.log_to_stdout(
          'Unrecognized input. Try again.',
          color=simulation_app.AnsiColors.YELLOW,
      )
      input_text = None
    else:
      return input_text == 'y'


def _toggle_log_progress(log_sim_progress: bool) -> bool:
  """Toggles the --log_progress flag."""
  log_sim_progress = not log_sim_progress
  simulation_app.log_to_stdout(
      f'--log_progress is now {log_sim_progress}.',
      color=simulation_app.AnsiColors.GREEN,
  )
  if log_sim_progress:
    simulation_app.log_to_stdout(
        'Each time step will be logged to stdout.',
        color=simulation_app.AnsiColors.GREEN,
    )
  else:
    simulation_app.log_to_stdout(
        'No time steps will be logged to stdout.',
        color=simulation_app.AnsiColors.GREEN,
    )
  return log_sim_progress


def _toggle_plot_progress(plot_sim_progress: bool) -> bool:
  """Toggles the --plot_progress flag."""
  plot_sim_progress = not plot_sim_progress
  simulation_app.log_to_stdout(
      f'--plot_progress is now {plot_sim_progress}.',
      color=simulation_app.AnsiColors.GREEN,
  )
  if plot_sim_progress:
    simulation_app.log_to_stdout(
        'Each time step will be shown in a plot.',
        color=simulation_app.AnsiColors.GREEN,
    )
  else:
    simulation_app.log_to_stdout(
        'No plots will be shown.',
        color=simulation_app.AnsiColors.GREEN,
    )
  return plot_sim_progress


def _toggle_log_output(log_sim_output: bool) -> bool:
  """Toggles the --log_output flag."""
  log_sim_output = not log_sim_output
  simulation_app.log_to_stdout(
      f'--log_output is now {log_sim_output}.',
      color=simulation_app.AnsiColors.GREEN,
  )
  if log_sim_output:
    simulation_app.log_to_stdout(
        'Extra simulation info will be logged to stdout at the end of the run.',
        color=simulation_app.AnsiColors.GREEN,
    )
  else:
    simulation_app.log_to_stdout(
        'No extra information will be logged to stdout.',
        color=simulation_app.AnsiColors.GREEN,
    )
  return log_sim_output


def _post_run_plotting(
    output_files: Sequence[str],
) -> None:
  """Helper to produce plots after a simulation run."""
  input_text = input(
      'Plot the last run (0), the last two runs (1), the last run against a'
      ' reference run (2): '
  )
  if not output_files:
    simulation_app.log_to_stdout(
        'No output files found, skipping plotting.',
        color=simulation_app.AnsiColors.RED,
    )
    return
  try:
    plot_config = config_loader.get_plot_config_from_file(
        _PLOT_CONFIG_PATH.value
    )
  except ValueError as e:
    logging.exception(
        'Error loading plot config module %s: %s', _PLOT_CONFIG_PATH.value, e
    )
    return
  match input_text:
    case '0':
      return plotruns_lib.plot_run(plot_config, output_files[-1])
    case '1':
      if len(output_files) == 1:
        simulation_app.log_to_stdout(
            'Only one output run file found, only plotting the last run.',
            color=simulation_app.AnsiColors.RED,
        )
        return plotruns_lib.plot_run(plot_config, output_files[-1])
      return plotruns_lib.plot_run(
          plot_config, output_files[-1], output_files[-2]
      )
    case '2':
      reference_run = _REFERENCE_RUN.value
      if reference_run is None:
        simulation_app.log_to_stdout(
            'No reference run provided, only plotting the last run.',
            color=simulation_app.AnsiColors.RED,
        )
      return plotruns_lib.plot_run(plot_config, output_files[-1], reference_run)
    case _:
      raise ValueError('Unknown command')


def main(_):
  torax.set_jax_precision()
  torax.log_jax_backend()

  if _CONFIG_PATH.value is None:
    raise ValueError(f'--{_CONFIG_PATH.name} must be specified.')

  config_path = pathlib.Path(_CONFIG_PATH.value)

  log_sim_progress = _LOG_SIM_PROGRESS.value
  plot_sim_progress = _PLOT_SIM_PROGRESS.value
  log_sim_output = _LOG_SIM_OUTPUT.value
  torax_config = None
  output_files = []
  output_dir = _OUTPUT_DIR.value if _OUTPUT_DIR is not None else None
  try:
    start_time = time.time()
    torax_config = config_loader.build_torax_config_from_file(config_path)
    build_time = time.time() - start_time
    start_time = time.time()
    output_file = _call_sim_app_main(
        torax_config,
        output_dir=output_dir,
        log_sim_progress=log_sim_progress,
        plot_sim_progress=plot_sim_progress,
        log_sim_output=log_sim_output,
    )
    simulation_time = time.time() - start_time
    output_files.append(output_file)
    simulation_app.log_to_stdout(
        f'Sim and params build time: {build_time:.2f}s, {SIMULATION_TIME}:'
        f' {simulation_time:.2f}s',
        color=simulation_app.AnsiColors.GREEN,
    )
  except ValueError as ve:
    simulation_app.log_to_stdout(
        f'Error occurred: {ve}',
        color=simulation_app.AnsiColors.RED,
        exc_info=True,
    )
    simulation_app.log_to_stdout(
        'Not running sim. Update config and try again.',
        color=simulation_app.AnsiColors.RED,
    )
  if _QUIT.value:
    return
  user_command = _prompt_user(config_path)
  while user_command != _UserCommand.QUIT:
    match user_command:
      case _UserCommand.QUIT:
        # This line shouldn't get hit, but is here for pytype.
        return  # Exit the function.
      case _UserCommand.RUN:
        if torax_config is None:
          simulation_app.log_to_stdout(
              'Need to reload the simulation.',
              color=simulation_app.AnsiColors.RED,
          )
        else:
          start_time = time.time()
          output_file = _call_sim_app_main(
              torax_config,
              output_dir=output_dir,
              log_sim_progress=log_sim_progress,
              plot_sim_progress=plot_sim_progress,
              log_sim_output=log_sim_output,
          )
          output_files.append(output_file)
          simulation_time = time.time() - start_time
          simulation_app.log_to_stdout(
              f'Simulation time: {simulation_time:.2f}s'
          )
      case _UserCommand.MODIFY_CONFIG:
        # See docstring for detailed info on what recompiles.
        if torax_config is None:
          simulation_app.log_to_stdout(
              'Need to reload the simulation.',
              color=simulation_app.AnsiColors.RED,
          )
        else:
          try:
            start_time = time.time()
            torax_config_or_none = _modify_config(config_path)
            if torax_config_or_none is not None:
              torax_config = torax_config_or_none
            config_change_time = time.time() - start_time
            simulation_app.log_to_stdout(
                f'Config change time: {config_change_time:.2f}s'
            )
          except ValueError as ve:
            simulation_app.log_to_stdout(
                f'Error occurred: {ve}',
                color=simulation_app.AnsiColors.RED,
            )
            simulation_app.log_to_stdout(
                'Update config and try again.',
                color=simulation_app.AnsiColors.RED,
            )
      case _UserCommand.CHANGE_CONFIG:
        # See docstring for detailed info on what recompiles.
        if torax_config is None:
          simulation_app.log_to_stdout(
              'Need to reload the simulation.',
              color=simulation_app.AnsiColors.RED,
          )
        else:
          try:
            start_time = time.time()
            torax_config_or_none = _change_config(config_path)
            if torax_config_or_none is not None:
              torax_config, config_path = torax_config_or_none
            config_change_time = time.time() - start_time
            simulation_app.log_to_stdout(
                f'Config change time: {config_change_time:.2f}s'
            )
          except ValueError as ve:
            simulation_app.log_to_stdout(
                f'Error occurred: {ve}',
                color=simulation_app.AnsiColors.RED,
            )
            simulation_app.log_to_stdout(
                'Update config and try again.',
                color=simulation_app.AnsiColors.RED,
            )
      case _UserCommand.TOGGLE_LOG_SIM_PROGRESS:
        log_sim_progress = _toggle_log_progress(log_sim_progress)
      case _UserCommand.TOGGLE_LOG_SIM_OUTPUT:
        log_sim_output = _toggle_log_output(log_sim_output)
      case _UserCommand.PLOT_RUN:
        _post_run_plotting(output_files)
      case _:
        raise ValueError('Unknown command')
    user_command = _prompt_user(config_path)


def use_jax_profiler_if_enabled(f):
  """Decorator that runs `func` with profiling if the flag is enabled."""

  @functools.wraps(f)
  def decorated(*args, **kwargs):
    if _USE_JAX_PROFILER.value:
      with jax.profiler.trace('/tmp/torax-jax-trace'):
        result = f(*args, **kwargs)
        simulation_app.log_to_stdout(
            'Profiling: to display results go to'
            ' https://ui.perfetto.dev/#!/viewer and open the trace file shown'
            ' below.'
        )
        return result
    else:
      return f(*args, **kwargs)

  return decorated


@use_jax_profiler_if_enabled
def _call_sim_app_main(
    torax_config,
    output_dir: str | None,
    log_sim_progress: bool,
    plot_sim_progress: bool,
    log_sim_output: bool,
):
  return simulation_app.main(
      lambda: torax_config,
      output_dir=output_dir,
      log_sim_progress=log_sim_progress,
      plot_sim_progress=plot_sim_progress,
      log_sim_output=log_sim_output,
  )


# Method used by the `run_torax` binary.
def run():
  app.run(main)


if __name__ == '__main__':
  run()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plasma composition parameters used throughout TORAX simulations."""
import copy
import dataclasses
import functools
import logging
from typing import Annotated, Any
import chex
import jax
import pydantic
from torax._src import array_typing
from torax._src.config import runtime_validation_utils
from torax._src.core_profiles.plasma_composition import electron_density_ratios
from torax._src.core_profiles.plasma_composition import electron_density_ratios_zeff
from torax._src.core_profiles.plasma_composition import impurity_fractions
from torax._src.core_profiles.plasma_composition import ion_mixture
from torax._src.torax_pydantic import torax_pydantic
import typing_extensions
from typing_extensions import Final

# pylint: disable=invalid-name

# Constants for impurity modes.
_IMPURITY_MODE_FRACTIONS: Final[str] = 'fractions'
_IMPURITY_MODE_NE_RATIOS: Final[str] = 'n_e_ratios'
_IMPURITY_MODE_NE_RATIOS_ZEFF: Final[str] = 'n_e_ratios_Z_eff'


@jax.tree_util.register_dataclass
@dataclasses.dataclass
class RuntimeParams:
  main_ion_names: tuple[str, ...] = dataclasses.field(metadata={'static': True})
  impurity_names: tuple[str, ...] = dataclasses.field(metadata={'static': True})
  main_ion: ion_mixture.RuntimeParams
  impurity: (
      ion_mixture.RuntimeParams
      | electron_density_ratios.RuntimeParams
      | electron_density_ratios_zeff.RuntimeParams
  )
  Z_eff: array_typing.FloatVectorCell
  Z_eff_face: array_typing.FloatVectorFace


# TODO(b/440667088): Consider validation against duplicate species.
@jax.tree_util.register_pytree_node_class
class PlasmaComposition(torax_pydantic.BaseModelFrozen):
  """Configuration for the plasma composition.

  The `from_dict(...)` method can accept a dictionary defined by
  https://torax.readthedocs.io/en/latest/configuration.html#plasma-composition.

  Attributes:
    main_ion: Main ion species. Can be single ion or a mixture of ions (e.g. D
      and T). Either a single ion, and constant mixture, or a time-dependent
      mixture. For single ions the input is one of the allowed strings in
      `ION_SYMBOLS`. For mixtures the input is an IonMixture object, constructed
      from a dict mapping ion symbols to their fractional concentration in the
      mixture.
    impurity: Dictionary configuring plasma impurities with the following keys:
      `impurity_mode`: Sets how the impurity species are defined. `species`: A
      dictionary mapping ion symbols (e.g., 'Ne', 'W') to their respective
      values. The interpretation of these values depends on the `impurity_mode`
      as follows. * `'fractions'`: relative fractional abundances. *
      `'n_e_ratios'`: ratios of impurity density to electron density. Z_eff is
      ignored in this mode. * `'n_e_ratios_Z_eff'`: ratios of impurity density
      to electron density. A single value must provided as None, and Z_eff is
      used to then constrain this value dynamically during runtime.
      `Z_override`: Optional. Overrides the calculated impurity average charge
      `A_override`: Optional. Overrides the calculated average impurity mass
      Backwards compatibility is provided for legacy inputs to `'impurity'`,
      e.g. string or dict inputs similar to `main_ion`, such as `'Ar'` or
      `{'Ar': 0.6, 'Ne': 0.4}`.
    Z_eff: Constraint for impurity densities. If the impurity_mode is
      `'n_e_ratios'`, and the input is not None, then any input value will be
      ignored and a warning provided to the user.
    Z_i_override: Optional arbitrary masses and charges which can be used to
      override the data for the average Z and A of each IonMixture for main_ions
      or impurities. Useful for testing or testing physical sensitivities,
      outside the constraint of allowed impurity species.
    A_i_override: Optional arbitrary masses and charges which can be used to
      override the data for the average Z and A of each IonMixture for main_ions
      or impurities. Useful for testing or testing physical sensitivities,
      outside the constraint of allowed impurity species.
    Z_impurity_override: DEPRECATED. As Z_i_override, but for the impurities.
    A_impurity_override: DEPRECATED. As A_i_override, but for the impurities.
  """

  impurity: Annotated[
      impurity_fractions.ImpurityFractions
      | electron_density_ratios.ElectronDensityRatios
      | electron_density_ratios_zeff.ElectronDensityRatiosZeff,
      pydantic.Field(discriminator='impurity_mode'),
  ]
  main_ion: runtime_validation_utils.IonMapping = (
      torax_pydantic.ValidatedDefault({'D': 0.5, 'T': 0.5})
  )
  Z_eff: (
      runtime_validation_utils.TimeVaryingArrayDefinedAtRightBoundaryAndBounded
  ) = torax_pydantic.ValidatedDefault(1.0)
  Z_i_override: torax_pydantic.TimeVaryingScalar | None = None
  A_i_override: torax_pydantic.TimeVaryingScalar | None = None
  Z_impurity_override: torax_pydantic.TimeVaryingScalar | None = None
  A_impurity_override: torax_pydantic.TimeVaryingScalar | None = None

  # For main_ions, IonMixture objects are generated by either a fractional
  # mixture (dict[str, TimeVaryingScalar]) or the shortcut for a single constant
  # ion (string).
  # For impurities, this input is legacy but still supported. A new API is also
  # available with different impurity_modes, e.g. fractions or n_e_ratios.
  # A pydantic before validator infers the API format and handles conversions.

  @pydantic.model_validator(mode='before')
  @classmethod
  def _conform_impurity_data(cls, data: dict[str, Any]) -> dict[str, Any]:
    """Sets defaults and ensures backward compatibility for impurity inputs."""
    configurable_data = copy.deepcopy(data)

    Z_impurity_override = configurable_data.get('Z_impurity_override')
    A_impurity_override = configurable_data.get('A_impurity_override')

    # Set defaults for impurity if not specified. To maintain same default
    # behaviour as before, the top-level Z_impurity_override and
    # A_impurity_override are used as overrides for the impurity fractions.
    # TODO(b/434175938): Remove this once V1 API is deprecated and the top-level
    # overrides are removed, and set default directly in class attribute.
    if 'impurity' not in configurable_data:
      configurable_data['impurity'] = {
          'impurity_mode': _IMPURITY_MODE_FRACTIONS,
          'Z_override': Z_impurity_override,
          'A_override': A_impurity_override,
          'legacy': True,
      }
      return configurable_data

    impurity_data = configurable_data['impurity']

    # New API format: impurity_mode is specified.
    if isinstance(impurity_data, dict) and 'impurity_mode' in impurity_data:
      if Z_impurity_override is not None or A_impurity_override is not None:
        logging.warning(
            'Z_impurity_override and/or A_impurity_override are set at the'
            ' plasma_composition level, but the new impurity API is being used'
            ' (impurity_mode is set). These top-level overrides are deprecated'
            ' and will be ignored. Use Z_override and A_override within the'
            ' impurity dictionary instead.'
        )
      return configurable_data

    # Legacy format from here on.
    # This handles conformant V1 inputs like 'Ne' or {'Ne': 0.8, 'Ar': 0.2}.
    # Non-conformant inputs are caught by ImpurityFractionsModel validation.
    # TODO(b/434175938): Remove this once V1 API is deprecated.
    configurable_data['impurity'] = {
        'impurity_mode': _IMPURITY_MODE_FRACTIONS,
        'species': impurity_data,
        'Z_override': Z_impurity_override,
        'A_override': A_impurity_override,
        'legacy': True,
    }
    return configurable_data

  @pydantic.model_validator(mode='after')
  def _check_zeff_usage(self) -> typing_extensions.Self:
    """Warns user if Z_eff is provided but will be ignored."""
    if (
        isinstance(self.impurity, electron_density_ratios.ElectronDensityRatios)
        and self.Z_eff.value != 1.0  # default value if input Z_eff is None
    ):
      logging.warning(
          "Z_eff is provided but impurity_mode is '%s'. Z_eff will be an"
          ' emergent quantity and the input value will be ignored.',
          _IMPURITY_MODE_NE_RATIOS,
      )
    return self

  def tree_flatten(self):
    # Override the default tree_flatten to also save out the cached
    # main_ion_mixture and impurity_mixture objects.
    children = (
        self.main_ion,
        self.impurity,
        self.Z_eff,
        self.Z_i_override,
        self.A_i_override,
        self.Z_impurity_override,
        self.A_impurity_override,
        self._main_ion_mixture,
    )
    aux_data = ()
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    obj = cls.model_construct(
        main_ion=children[0],
        impurity=children[1],
        Z_eff=children[2],
        Z_i_override=children[3],
        A_i_override=children[4],
        Z_impurity_override=children[5],
        A_impurity_override=children[6],
    )
    obj._main_ion_mixture = children[7]  # pylint: disable=protected-access
    return obj

  @functools.cached_property
  def _main_ion_mixture(self) -> ion_mixture.IonMixture:
    """Returns the IonMixture object for the main ions."""
    # Use `model_construct` as no validation required.
    return ion_mixture.IonMixture.model_construct(
        species=self.main_ion,
        Z_override=self.Z_i_override,
        A_override=self.A_i_override,
    )

  def get_main_ion_names(self) -> tuple[str, ...]:
    """Returns the main ion symbol strings from the input."""
    return tuple(self._main_ion_mixture.species.keys())

  def get_impurity_names(self) -> tuple[str, ...]:
    """Returns the impurity symbol strings from the input."""
    return tuple(self.impurity.species.keys())

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    return RuntimeParams(
        main_ion_names=self.get_main_ion_names(),
        impurity_names=self.get_impurity_names(),
        main_ion=self._main_ion_mixture.build_runtime_params(t),
        impurity=self.impurity.build_runtime_params(t),
        Z_eff=self.Z_eff.get_value(t),
        Z_eff_face=self.Z_eff.get_value(t, grid_type='face'),
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numerics parameters used throughout TORAX simulations."""

import dataclasses
import functools
from typing import Annotated

import chex
import jax
import pydantic
from torax._src import array_typing
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import Self


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass
class RuntimeParams:
  """Generic numeric parameters for the simulation.

  For definitions see `Numerics`.
  """

  t_initial: float
  t_final: float
  max_dt: float
  min_dt: float
  chi_timestep_prefactor: float
  fixed_dt: float
  dt_reduction_factor: float
  resistivity_multiplier: array_typing.FloatScalar
  adaptive_T_source_prefactor: float
  adaptive_n_source_prefactor: float
  evolve_ion_heat: bool = dataclasses.field(metadata={'static': True})
  evolve_electron_heat: bool = dataclasses.field(metadata={'static': True})
  evolve_current: bool = dataclasses.field(metadata={'static': True})
  evolve_density: bool = dataclasses.field(metadata={'static': True})
  exact_t_final: bool = dataclasses.field(metadata={'static': True})
  adaptive_dt: bool = dataclasses.field(metadata={'static': True})
  calcphibdot: bool = dataclasses.field(metadata={'static': True})

  @functools.cached_property
  def evolving_names(self) -> tuple[str, ...]:
    """The names of core_profiles variables that are evolved by the solver."""
    evolving_names = []
    if self.evolve_ion_heat:
      evolving_names.append('T_i')
    if self.evolve_electron_heat:
      evolving_names.append('T_e')
    if self.evolve_current:
      evolving_names.append('psi')
    if self.evolve_density:
      evolving_names.append('n_e')
    return tuple(evolving_names)


class Numerics(torax_pydantic.BaseModelFrozen):
  """Generic numeric parameters for the simulation.

  The `from_dict(...)` method can accept a dictionary defined by
  https://torax.readthedocs.io/en/latest/configuration.html#numerics.

  Attributes:
    t_initial: Simulation start time, in units of seconds.
    t_final: Simulation end time, in units of seconds.
    exact_t_final: If True, ensures that the simulation end time is exactly
      `t_final`, by adapting the final `dt` to match.
    max_dt: Maximum timesteps allowed in the simulation. This is only used with
      the `chi_time_step_calculator` time_step_calculator.
    min_dt: Minimum timestep allowed in simulation.
    chi_timestep_prefactor: Prefactor in front of chi_timestep_calculator base
      timestep dt=dx^2/(2*chi). In most use-cases can be increased further above
      this.
    fixed_dt: Timestep used for `fixed_time_step_calculator`.
    adaptive_dt: Iterative reduction of dt if nonlinear step does not converge,
      if nonlinear step does not converge, then the step is redone iteratively
      at successively lower dt until convergence is reached.
    dt_reduction_factor: Factor by which to reduce dt if adaptive_dt is True.
    evolve_ion_heat: Solve the ion heat equation (ion temperature evolves over
      time).
    evolve_electron_heat: Solve the electron heat equation (electron temperature
      evolves over time)
    evolve_current: Solve the current equation (current evolves over time).
    evolve_density: Solve the density equation (n evolves over time).
    calcphibdot: Calculate Phibdot in the geometry dataclasses. This is used in
      calc_coeffs to calculate terms related to time-dependent geometry. Can set
      to false to zero out for testing purposes.
    resistivity_multiplier:  1/multiplication factor for sigma (conductivity) to
      reduce current diffusion timescale to be closer to heat diffusion
      timescale
    adaptive_T_source_prefactor: Prefactor for adaptive source term for setting
      temperature internal boundary conditions.
    adaptive_n_source_prefactor: Prefactor for adaptive source term for setting
      density internal boundary conditions.
  """

  t_initial: torax_pydantic.Second = 0.0
  t_final: torax_pydantic.Second = 5.0
  exact_t_final: Annotated[bool, torax_pydantic.JAX_STATIC] = True
  max_dt: torax_pydantic.Second = 2.0
  min_dt: torax_pydantic.Second = 1e-8
  chi_timestep_prefactor: pydantic.PositiveFloat = 50.0
  fixed_dt: torax_pydantic.Second = 1e-1
  adaptive_dt: Annotated[bool, torax_pydantic.JAX_STATIC] = True
  dt_reduction_factor: pydantic.PositiveFloat = 3.0
  evolve_ion_heat: Annotated[bool, torax_pydantic.JAX_STATIC] = True
  evolve_electron_heat: Annotated[bool, torax_pydantic.JAX_STATIC] = True
  evolve_current: Annotated[bool, torax_pydantic.JAX_STATIC] = False
  evolve_density: Annotated[bool, torax_pydantic.JAX_STATIC] = False
  calcphibdot: Annotated[bool, torax_pydantic.JAX_STATIC] = True
  resistivity_multiplier: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(1.0)
  )
  adaptive_T_source_prefactor: pydantic.PositiveFloat = 2.0e10
  adaptive_n_source_prefactor: pydantic.PositiveFloat = 2.0e8

  @pydantic.model_validator(mode='after')
  def model_validation(self) -> Self:
    if self.t_initial > self.t_final:
      raise ValueError(
          't_initial must be less than or equal to t_final. '
          f't_initial: {self.t_initial}, t_final: {self.t_final}'
      )

    if self.min_dt > self.max_dt:
      raise ValueError(
          'max_dt must be greater than or equal to min_dt. '
          f'max_dt: {self.max_dt}, min_dt: {self.min_dt}'
      )
    return self

  @property
  def evolving_names(self) -> tuple[str, ...]:
    """The names of core_profiles variables that are evolved by the solver."""
    evolving_names = []
    if self.evolve_ion_heat:
      evolving_names.append('T_i')
    if self.evolve_electron_heat:
      evolving_names.append('T_e')
    if self.evolve_current:
      evolving_names.append('psi')
    if self.evolve_density:
      evolving_names.append('n_e')
    return tuple(evolving_names)

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    """Builds a RuntimeParams object for time t."""
    return RuntimeParams(
        t_initial=self.t_initial,
        t_final=self.t_final,
        max_dt=self.max_dt,
        min_dt=self.min_dt,
        chi_timestep_prefactor=self.chi_timestep_prefactor,
        fixed_dt=self.fixed_dt,
        dt_reduction_factor=self.dt_reduction_factor,
        resistivity_multiplier=self.resistivity_multiplier.get_value(t),
        adaptive_T_source_prefactor=self.adaptive_T_source_prefactor,
        adaptive_n_source_prefactor=self.adaptive_n_source_prefactor,
        evolve_ion_heat=self.evolve_ion_heat,
        evolve_electron_heat=self.evolve_electron_heat,
        evolve_current=self.evolve_current,
        evolve_density=self.evolve_density,
        exact_t_final=self.exact_t_final,
        adaptive_dt=self.adaptive_dt,
        calcphibdot=self.calcphibdot,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for MHD models."""

import dataclasses

import jax
from torax._src.mhd.sawtooth import sawtooth_models as sawtooth_models_lib


@jax.tree_util.register_dataclass
@dataclasses.dataclass
class MHDModels:
  """Container for instantiated MHD model objects."""

  sawtooth_models: sawtooth_models_lib.SawtoothModels | None = None

  def __eq__(self, other: 'MHDModels') -> bool:
    return self.sawtooth_models == other.sawtooth_models

  def __hash__(self) -> int:
    return hash((self.sawtooth_models,))



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic config for restarting a simulation from a file."""
import pydantic
from torax._src.torax_pydantic import torax_pydantic


class FileRestart(torax_pydantic.BaseModelFrozen):
  """Pydantic config for restarting a simulation from a file.

  Attributes:
    filename: Path to an existing TORAX output file (netCDF format).
    time: Time in output file to load as the new initial state.
    do_restart: If True, perform the restart from the selected state.
      If False, disable the restart and run the simulation as normal.
    stitch: If True, concatenate the old and new simulation histories
      in the resulting output file. If False, output file will only
      contain the new history.
  """

  filename: pydantic.FilePath
  time: torax_pydantic.Second
  do_restart: bool
  stitch: bool



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Commonly repeated jax expressions."""

import contextlib
import functools
import inspect
import os
from typing import Any, Callable, Literal, ParamSpec, TypeAlias, TypeVar

import chex
import equinox as eqx
import jax
from jax import numpy as jnp
import numpy as np

T = TypeVar('T')
BooleanNumeric: TypeAlias = Any  # A bool, or a Boolean array.
_State = ParamSpec('_State')


@functools.cache
def get_dtype() -> type(jnp.float32):
  # Default TORAX JAX precision is f64
  precision = os.getenv('JAX_PRECISION', 'f64')
  assert precision == 'f64' or precision == 'f32', (
      'Unknown JAX precision environment variable: %s' % precision
  )
  return jnp.float64 if precision == 'f64' else jnp.float32


@functools.cache
def get_np_dtype() -> type(np.float32):
  # Default TORAX JAX precision is f64
  precision = os.getenv('JAX_PRECISION', 'f64')
  assert precision == 'f64' or precision == 'f32', (
      'Unknown JAX precision environment variable: %s' % precision
  )
  return np.float64 if precision == 'f64' else np.float32


@functools.cache
def get_int_dtype() -> type(jnp.int32):
  # Default TORAX JAX precision is f64
  precision = os.getenv('JAX_PRECISION', 'f64')
  assert precision == 'f64' or precision == 'f32', (
      'Unknown JAX precision environment variable: %s' % precision
  )
  return jnp.int64 if precision == 'f64' else jnp.int32


def env_bool(name: str, default: bool) -> bool:
  """Get a bool from an environment variable.

  Args:
    name: The name of the environment variable.
    default: The default value of the bool.

  Returns:
    value: The value of the bool.
  """
  if name not in os.environ:
    return default
  str_value = os.environ[name]
  if str_value in ['1', 'True', 'true']:
    return True
  if str_value in ['0', 'False', 'false']:
    return False
  raise ValueError(f'Unrecognized boolean string {str_value}.')


# If True, `error_if` functions will raise errors.  Otherwise they are
# pass throughs.
# Default to False, because host_callbacks are incompatible with the
# persistent compilation cache.
_ERRORS_ENABLED: bool = env_bool('TORAX_ERRORS_ENABLED', False)


@contextlib.contextmanager
def enable_errors(value: bool):
  """Enables / disables `error_if` inside a code block.

  Example:

  with enable_errors(False):
    my_sim.run() # NaNs etc will be ignored

  Args:
    value: Sets `errors_enabled` to this value

  Yields:
    Cleanup function restoring previous value
  """
  global _ERRORS_ENABLED
  previous_value = _ERRORS_ENABLED
  _ERRORS_ENABLED = value
  yield
  if previous_value is not None:
    _ERRORS_ENABLED = previous_value


def error_if(
    var: jax.Array,
    cond: jax.Array,
    msg: str,
) -> jax.Array:
  """Raises error if cond is true, and `errors_enabled` is True.

  This is just a wrapper around `equinox.error_if`, gated by `errors_enabled`.

  Args:
    var: The variable to pass through.
    cond: Boolean array, error if cond is true.
    msg: Message to print on error.

  Returns:
    var: Identity wrapper that must be used for the check to be included.
  """
  if not _ERRORS_ENABLED:
    return var
  return eqx.error_if(var, cond, msg)


def assert_rank(
    inputs: chex.Numeric | jax.stages.ArgInfo,
    rank: int,
) -> None:
  """Wrapper around chex.assert_rank that supports jax.stages.ArgInfo."""
  if isinstance(inputs, jax.stages.ArgInfo):
    chex.assert_rank(inputs.shape, rank)
  else:
    chex.assert_rank(inputs, rank)


def jit(*args, **kwargs) -> Callable[..., Any]:
  """Calls jax.jit if TORAX_COMPILATION_ENABLED is True, otherwise no-op."""
  if env_bool('TORAX_COMPILATION_ENABLED', True):
    return jax.jit(*args, **kwargs)
  return args[0]


def get_number_of_compiles(
    jitted_function: Callable[..., Any],
) -> int:
  """Helper function for debugging JAX compilation.

  This counts the number of times the function has been JIT compiled. This does
  not include any uses of the AOT compile workflow.

  Args:
    jitted_function: A function that has been wrapped with `jax.jit`.

  Returns:
    The number of times the function has been compiled.
  Raises:
    RuntimeError: If the function does not have a _cache_size attribute.
  """
  # pylint: disable=protected-access
  if not hasattr(jitted_function, '_cache_size'):
    raise RuntimeError(
        'The function does not have a _cache_size attribute. Possibly because'
        ' the function was not jitted.'
    )
  return jitted_function._cache_size()
  # pylint: enable=protected-access


# pylint: enable=g-bare-generic


# TODO(b/424382924)
def non_inlined_function(
    f: Callable[..., Any],
    implementation: Literal['while_loop', 'pure_callback'],
) -> Callable[..., Any]:
  """A decorator that prevents XLA from inlining a function.

  XLA inlines all functions in a computational graph. As XLA does global
  optimization, the compile times increase super-linearly. This decorator
  allows preventing inlining of functions using `jax.lax.while_loop` or
  `jax.pure_callback`. In the case of `jax.pure_callback`, what is called from
  the Python callback is a black box to XLA, and cannot be inlined. In the case
  of `jax.lax.while_loop`, the body function is not inlined with the rest of the
  computation.

  Args:
    f: A JITted function.
    implementation: If 'while_loop', use `jax.lax.while_loop` with a single
      iteration. If 'pure_callback', use `jax.pure_callback`. This comes at the
      cost of a roughly 0.7ms constant overhead per call. It is recommended that
      `f` is a JITted function in this case, as it will be called directly from
      Python.

  Returns:
    The function.
  """

  if not hasattr(f, 'lower'):
    raise ValueError('Must be a JITted function.')

  if not hasattr(f, '_jit_info'):
    raise ValueError('The function must have a _jit_info attribute.')

  static_argnames = f._jit_info.static_argnames  # pylint: disable=protected-access

  match implementation:
    case 'while_loop':
      return _non_inlined_function_while_loop(
          f, static_argnames=static_argnames
      )
    case 'pure_callback':
      return _non_inlined_function_pure_callback(
          f, static_argnames=static_argnames
      )
    case _:
      raise ValueError(f'Unknown implementation: {implementation}')


def _non_inlined_function_pure_callback(f, static_argnames):
  """A decorator that prevents XLA from inlining a function."""

  @functools.wraps(f)
  def wrapper(*args, **kwargs):
    nonlocal f, static_argnames
    bound = inspect.signature(f).bind(*args, **kwargs)
    bound.apply_defaults()
    kwargs = bound.arguments
    if 'self' in kwargs:
      kwargs.pop('self')

    if static_argnames:
      static_args = {k: bound.arguments[k] for k in static_argnames}

      kwargs = {k: v for k, v in kwargs.items() if k not in static_argnames}
      f = functools.partial(f, **static_args)

    result_shape_dtypes = jax.eval_shape(f, **kwargs)
    return jax.pure_callback(f, result_shape_dtypes, **kwargs)

  return wrapper


def _non_inlined_function_while_loop(f, static_argnames):
  """A decorator that prevents XLA from inlining a function."""

  @functools.wraps(f)
  def wrapper(*args, **kwargs):
    nonlocal f, static_argnames
    bound = inspect.signature(f).bind(*args, **kwargs)
    bound.apply_defaults()
    kwargs = bound.arguments
    if 'self' in kwargs:
      kwargs.pop('self')

    if static_argnames:
      static_args = {k: bound.arguments[k] for k in static_argnames}
      kwargs = {k: v for k, v in kwargs.items() if k not in static_argnames}
      f = functools.partial(f, **static_args)

    continue_loop = True
    empty_out = _init_pytree(jax.eval_shape(f, **kwargs))

    def body(val):
      return False, val[1], f(**val[1])

    def cond(val):
      return val[0]

    out = jax.lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=(continue_loop, kwargs, empty_out),
    )
    return out[-1]

  return wrapper


def _init_pytree(t):

  def init_array(x):
    if isinstance(x, jax.ShapeDtypeStruct):
      return jnp.empty(shape=x.shape, dtype=x.dtype)
    else:
      return x

  return jax.tree_util.tree_map(init_array, t)


@functools.partial(
    jit, static_argnames=['cond_fun', 'body_fun', 'max_steps', 'scan_unroll']
)
def while_loop_bounded(
    cond_fun: Callable[[_State], BooleanNumeric],
    body_fun: Callable[[_State], _State],
    init_val: _State,
    max_steps: int,
    scan_unroll: int = 1,
) -> _State:
  """A reverse-mode differentiable while_loop.

  This makes use of jax.lax.scan and `max_steps` to define a fixed size
  computational graph. The body_fun is called the same number of times it would
  be under a jax.lax.while_loop i.e. until `cond_fun` returns False (unless the
  `max_steps` is reached).

  Args:
    cond_fun: As in jax.lax.while_loop.
    body_fun: As in jax.lax.while_loop.
    init_val: As in jax.lax.while_loop.
    max_steps: An integer, the maximum number of iterations the loop can
      perform. This is crucial for defining a fixed computational graph for
      scan.
    scan_unroll: The number of iterations to unroll the internal scan by.

  Returns:
    The final state after `cond_fun` returns `False` or `max_steps` are reached.
  """
  # Initial carry for the scan: (current_state, while_loop_condition_met)
  initial_scan_carry = (init_val, jnp.array(True, dtype=jnp.bool_))

  def scan_body(carry, _):
    current_state, cond_prev = carry
    # Only execute cond if the previous cond was True.
    should_execute_body = jax.lax.cond(
        cond_prev, cond_fun, lambda _: False, current_state
    )
    # If the `while_loop` would have terminated, we no-op.
    next_state = jax.lax.cond(
        should_execute_body, body_fun, lambda s: s, current_state
    )

    return (next_state, should_execute_body), None

  (final_state, _), _ = jax.lax.scan(
      scan_body, initial_scan_carry, length=max_steps, unroll=scan_unroll
  )

  return final_state



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Inputs to TORAX solvers and functions based on the input runtime parameters.

When running a TORAX simulation, the solver is (by default) a JAX-compiled
function, meaning it has two types of arguments: "dynamic" and "static".

The "dynamic" arguments can change from call to call. These arguments must be
arrays, scalars, or standard (possibly nested) Python containers. See the JAX
docs for more info on allowed types. They cannot influence the logical branches
the SimulationStepFn may take (again, see the sharp bits in the JAX docs to
learn more about the how these "dynamic" args can be used within the function).
By default all the arguments in runtime_params are assumed to be dynamic.

Note that the "dynamic" arguments are NOT necessarily time-dependent. They do
not need to vary from time step to time step (though they can). They can change
from time step to time step, or from simulation run to simulation run, without
triggering a recompile. Changing these params without needing to recompile the
solver is the defining quality of the dynamic arguments.

The "static" arguments are compile-time constant. Any changes to them would
trigger a recompilation of the solver. These arguments don't have the same
restrictions as the dynamic arguments both in terms of types and how they are
used. They are marked as static in their definition and are marked as such when
input to JAX.
"""
from collections.abc import Mapping
import dataclasses

import jax
from torax._src.config import numerics
from torax._src.core_profiles import profile_conditions
from torax._src.core_profiles.plasma_composition import plasma_composition
from torax._src.geometry import geometry
from torax._src.geometry import standard_geometry
from torax._src.mhd import runtime_params as mhd_runtime_params
from torax._src.neoclassical import runtime_params as neoclassical_params
from torax._src.pedestal_model import runtime_params as pedestal_model_params
from torax._src.solver import runtime_params as solver_params
from torax._src.sources import runtime_params as sources_params
from torax._src.time_step_calculator import runtime_params as time_step_calculator_runtime_params
from torax._src.transport_model import runtime_params as transport_model_params

# Many of the variables follow scientific or mathematical notation, so disable
# pylint complaints.
# pylint: disable=invalid-name


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams:
  """A slice of the parameters at a specific time t.

  This PyTree is a slice of the overall TORAX config at a specific time t
  excluding the geometry, grouping the parameters for ease of passing into
  various downstream functions. It includes both parameters which are
  time-dependent and parameters which are not.

  The parameters which are not time-dependent are marked as static in their
  definition, this means that they are marked as static when input to JAX,
  which means that they are compile-time constants. This means that they cannot
  be changed without recompilation.
  """

  mhd: mhd_runtime_params.RuntimeParams
  neoclassical: neoclassical_params.RuntimeParams
  numerics: numerics.RuntimeParams
  pedestal: pedestal_model_params.RuntimeParams
  plasma_composition: plasma_composition.RuntimeParams
  profile_conditions: profile_conditions.RuntimeParams
  solver: solver_params.RuntimeParams
  sources: Mapping[str, sources_params.RuntimeParams]
  transport: transport_model_params.RuntimeParams
  time_step_calculator: time_step_calculator_runtime_params.RuntimeParams


def make_ip_consistent(
    runtime_params: RuntimeParams,
    geo: geometry.Geometry,
) -> tuple[RuntimeParams, geometry.Geometry]:
  """Fixes Ip to be the same across runtime_params and geo."""
  if isinstance(geo, standard_geometry.StandardGeometry):
    if geo.Ip_from_parameters:
      # If Ip is from parameters, renormalise psi etc to match the Ip in the
      # parameters.
      param_Ip = runtime_params.profile_conditions.Ip
      Ip_scale_factor = param_Ip / geo.Ip_profile_face[-1]
      geo = dataclasses.replace(
          geo,
          Ip_profile_face=geo.Ip_profile_face * Ip_scale_factor,
          psi_from_Ip=geo.psi_from_Ip * Ip_scale_factor,
          psi_from_Ip_face=geo.psi_from_Ip_face * Ip_scale_factor,
          j_total=geo.j_total * Ip_scale_factor,
          j_total_face=geo.j_total_face * Ip_scale_factor,
      )
    else:
      # If Ip is from the geometry, update the parameters to match.
      runtime_params = dataclasses.replace(
          runtime_params,
          profile_conditions=dataclasses.replace(
              runtime_params.profile_conditions,
              Ip=geo.Ip_profile_face[-1],
          ),
      )
  return runtime_params, geo



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic config for Pedestal."""
import abc
from typing import Annotated, Literal

import chex
from torax._src.pedestal_model import no_pedestal
from torax._src.pedestal_model import pedestal_model
from torax._src.pedestal_model import runtime_params
from torax._src.pedestal_model import set_pped_tpedratio_nped
from torax._src.pedestal_model import set_tped_nped
from torax._src.torax_pydantic import torax_pydantic

# pylint: disable=invalid-name


class BasePedestal(torax_pydantic.BaseModelFrozen, abc.ABC):
  """Base class for pedestal models.

  Attributes:
    set_pedestal: Whether to use the pedestal model and set the pedestal. Can be
      time varying.
  """

  set_pedestal: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(False)
  )

  @abc.abstractmethod
  def build_pedestal_model(self) -> pedestal_model.PedestalModel:
    """Builds the pedestal model."""

  @abc.abstractmethod
  def build_runtime_params(
      self, t: chex.Numeric
  ) -> runtime_params.RuntimeParams:
    """Builds the runtime params."""


class SetPpedTpedRatioNped(BasePedestal):
  """Model for direct specification of pressure, temperature ratio, and density.

  Attributes:
    P_ped: The plasma pressure at the pedestal [Pa].
    n_e_ped: The electron density at the pedestal [m^-3] or fGW.
    n_e_ped_is_fGW: Whether the electron density at the pedestal is in units of
      fGW.
    T_i_T_e_ratio: Ratio of the ion and electron temperature at the pedestal
      [dimensionless].
    rho_norm_ped_top: The location of the pedestal top.
  """

  model_name: Annotated[
      Literal['set_P_ped_n_ped'], torax_pydantic.JAX_STATIC
  ] = 'set_P_ped_n_ped'
  P_ped: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(1e5)
  n_e_ped: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      0.7e20
  )
  n_e_ped_is_fGW: bool = False
  T_i_T_e_ratio: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(1.0)
  )
  rho_norm_ped_top: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.91)
  )

  def build_pedestal_model(
      self,
  ) -> (
      set_pped_tpedratio_nped.SetPressureTemperatureRatioAndDensityPedestalModel
  ):
    return (
        set_pped_tpedratio_nped.SetPressureTemperatureRatioAndDensityPedestalModel()
    )

  def build_runtime_params(
      self, t: chex.Numeric
  ) -> set_pped_tpedratio_nped.RuntimeParams:
    return set_pped_tpedratio_nped.RuntimeParams(
        set_pedestal=self.set_pedestal.get_value(t),
        P_ped=self.P_ped.get_value(t),
        n_e_ped=self.n_e_ped.get_value(t),
        n_e_ped_is_fGW=self.n_e_ped_is_fGW,
        T_i_T_e_ratio=self.T_i_T_e_ratio.get_value(t),
        rho_norm_ped_top=self.rho_norm_ped_top.get_value(t),
    )


class SetTpedNped(BasePedestal):
  """A basic version of the pedestal model that uses direct specification.

  Attributes:
    n_e_ped: The electron density at the pedestal [m^-3] or fGW.
    n_e_ped_is_fGW: Whether the electron density at the pedestal is in units of
      fGW.
    T_i_ped: Ion temperature at the pedestal [keV].
    T_e_ped: Electron temperature at the pedestal [keV].
    rho_norm_ped_top: The location of the pedestal top.
  """

  model_name: Annotated[
      Literal['set_T_ped_n_ped'], torax_pydantic.JAX_STATIC
  ] = 'set_T_ped_n_ped'
  n_e_ped: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      0.7e20
  )
  n_e_ped_is_fGW: bool = False
  T_i_ped: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      5.0
  )
  T_e_ped: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      5.0
  )
  rho_norm_ped_top: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.91)
  )

  def build_pedestal_model(
      self,
  ) -> set_tped_nped.SetTemperatureDensityPedestalModel:
    return set_tped_nped.SetTemperatureDensityPedestalModel()

  def build_runtime_params(
      self, t: chex.Numeric
  ) -> set_tped_nped.RuntimeParams:
    return set_tped_nped.RuntimeParams(
        set_pedestal=self.set_pedestal.get_value(t),
        n_e_ped=self.n_e_ped.get_value(t),
        n_e_ped_is_fGW=self.n_e_ped_is_fGW,
        T_i_ped=self.T_i_ped.get_value(t),
        T_e_ped=self.T_e_ped.get_value(t),
        rho_norm_ped_top=self.rho_norm_ped_top.get_value(t),
    )


class NoPedestal(BasePedestal):
  """A pedestal model for when there is no pedestal.

  Note that setting `set_pedestal` to True with a NoPedestal model is the
  equivalent of setting it to False.
  """

  model_name: Annotated[Literal['no_pedestal'], torax_pydantic.JAX_STATIC] = (
      'no_pedestal'
  )

  def build_pedestal_model(
      self,
  ) -> no_pedestal.NoPedestal:
    return no_pedestal.NoPedestal()

  def build_runtime_params(
      self, t: chex.Numeric
  ) -> runtime_params.RuntimeParams:
    return runtime_params.RuntimeParams(
        set_pedestal=self.set_pedestal.get_value(t),
    )


PedestalConfig = SetPpedTpedRatioNped | SetTpedNped | NoPedestal



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Container for source models which build source profiles in TORAX."""
import dataclasses
import functools

import immutabledict
import jax
from torax._src.sources import qei_source as qei_source_lib
from torax._src.sources import source as source_lib


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class SourceModels:
  """Source models for the different equations being evolved in Torax."""

  qei_source: qei_source_lib.QeiSource
  standard_sources: immutabledict.immutabledict[str, source_lib.Source]

  @functools.cached_property
  def psi_sources(self) -> immutabledict.immutabledict[str, source_lib.Source]:
    """A derived dictionary of sources that affect the psi core_profile."""
    return immutabledict.immutabledict({
        name: source
        for name, source in self.standard_sources.items()
        if source_lib.AffectedCoreProfile.PSI in source.affected_core_profiles
    })

  def __hash__(self) -> int:
    hashes = [hash(self.standard_sources)]
    hashes.append(hash(self.qei_source))
    return hash(tuple(hashes))

  def __eq__(self, other) -> bool:
    if set(self.standard_sources.keys()) == set(other.standard_sources.keys()):
      return (
          all(
              self.standard_sources[name] == other.standard_sources[name]
              for name in self.standard_sources.keys()
          )
          and self.qei_source == other.qei_source
      )
    return False
